//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include <windows.h>
#include <stdio.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"
#include <math.h>
#include "resample.h"

#include "cpuaccel.h"

#define USE_ASM

static void MakeCubic4Table(int *table, double A, bool mmx_table) {
	int i;

	for(i=0; i<256; i++) {
		double d = (double)i / 256.0;
		int y1, y2, y3, y4, ydiff;

		// Coefficients for all four pixels *must* add up to 1.0 for
		// consistent unity gain.
		//
		// Two good values for A are -1.0 (original VirtualDub bicubic filter)
		// and -0.75 (closely matches Photoshop).

		y1 = (int)(0.5 + (        +     A*d -       2.0*A*d*d +       A*d*d*d) * 16384.0);
		y2 = (int)(0.5 + (+ 1.0             -     (A+3.0)*d*d + (A+2.0)*d*d*d) * 16384.0);
		y3 = (int)(0.5 + (        -     A*d + (2.0*A+3.0)*d*d - (A+2.0)*d*d*d) * 16384.0);
		y4 = (int)(0.5 + (                  +           A*d*d -       A*d*d*d) * 16384.0);

		// Normalize y's so they add up to 16384.

		ydiff = (16384 - y1 - y2 - y3 - y4)/4;

		y1 += ydiff;
		y2 += ydiff;
		y3 += ydiff;
		y4 += ydiff;

		if (mmx_table) {
			table[i*4 + 0] = table[i*4 + 1] = (y2<<16) | (y1 & 0xffff);
			table[i*4 + 2] = table[i*4 + 3] = (y4<<16) | (y3 & 0xffff);
		} else {
			table[i*4 + 0] = y1;
			table[i*4 + 1] = y2;
			table[i*4 + 2] = y3;
			table[i*4 + 3] = y4;
		}
	}
}

/////////////////////////////////////////////////////////////////////

typedef struct RotateRow {
	int leftzero, left, right, rightzero;
	__int64 xaccum_left, yaccum_left;
} RotateRow;

typedef struct MyFilterData2 {
	int angle;
	int filtmode;

	// working variables

	IFilterPreview *ifp;

	__int64	u_step;
	__int64	v_step;

	RotateRow *rows;
	int *coeff_tbl;

	COLORREF	rgbColor;
	HBRUSH		hbrColor;

	bool	fExpandBounds;
	
} MyFilterData2;

enum {
	FILTMODE_POINT		= 0,
	FILTMODE_BILINEAR,
	FILTMODE_BICUBIC,
	FILTMODE_COUNT
};

/////////////////////////////////////////////////////////////////////

extern "C" void asm_rotate_point(
		Pixel *src,
		Pixel *dst,
		long width,
		long Ufrac,
		long Vfrac,
		long UVintstepV,
		long UVintstepnoV,
		long Ustep,
		long Vstep);

extern "C" void asm_rotate_bilinear(
		Pixel *src,
		Pixel *dst,
		long width,
		long pitch,
		long Ufrac,
		long Vfrac,
		long UVintstepV,
		long UVintstepnoV,
		long Ustep,
		long Vstep);

/////////////////////////////////////////////////////////////////////

static Pixel32 bilinear_interp(Pixel32 c1, Pixel32 c2, Pixel32 c3, Pixel32 c4, unsigned long cox, unsigned long coy) {
	int co1, co2, co3, co4;

	co4 = (cox * coy) >> 8;
	co3 = coy - co4;
	co2 = cox - co4;
	co1 = 0x100 - coy - co2;

	return   ((((c1 & 0x00FF00FF)*co1 + (c2 & 0x00FF00FF)*co2 + (c3 & 0x00FF00FF)*co3 + (c4 & 0x00FF00FF)*co4)>>8)&0x00FF00FF)
		   + ((((c1 & 0x0000FF00)*co1 + (c2 & 0x0000FF00)*co2 + (c3 & 0x0000FF00)*co3 + (c4 & 0x0000FF00)*co4)&0x00FF0000)>>8);
}

#define RED(x) ((signed long)((x)>>16)&255)
#define GRN(x) ((signed long)((x)>> 8)&255)
#define BLU(x) ((signed long)(x)&255)

static inline Pixel cc(const Pixel *yptr, const int *tbl) {
	const Pixel y1 = yptr[0];
	const Pixel y2 = yptr[1];
	const Pixel y3 = yptr[2];
	const Pixel y4 = yptr[3];
	long red, grn, blu;

	red = RED(y1)*tbl[0] + RED(y2)*tbl[1] + RED(y3)*tbl[2] + RED(y4)*tbl[3];
	grn = GRN(y1)*tbl[0] + GRN(y2)*tbl[1] + GRN(y3)*tbl[2] + GRN(y4)*tbl[3];
	blu = BLU(y1)*tbl[0] + BLU(y2)*tbl[1] + BLU(y3)*tbl[2] + BLU(y4)*tbl[3];

	if (red<0) red=0; else if (red>4194303) red=4194303;
	if (grn<0) grn=0; else if (grn>4194303) grn=4194303;
	if (blu<0) blu=0; else if (blu>4194303) blu=4194303;

	return ((red<<2) & 0xFF0000) | ((grn>>6) & 0x00FF00) | (blu>>14);
}

#undef RED
#undef GRN
#undef BLU

static Pixel32 __declspec(naked) cc_MMX(const Pixel32 *src, const int *table) {

	static const __int64 x0000200000002000 = 0x0000200000002000i64;

	//	[esp + 4]	src
	//	[esp + 8]	table

	_asm {
		mov			ecx,[esp+4]
		mov			eax,[esp+8]

		movd		mm0,[ecx]
		pxor		mm7,mm7

		movd		mm1,[ecx+4]
		punpcklbw	mm0,mm7				;mm0 = [a1][r1][g1][b1]

		movd		mm2,[ecx+8]
		punpcklbw	mm1,mm7				;mm1 = [a2][r2][g2][b2]

		movd		mm3,[ecx+12]
		punpcklbw	mm2,mm7				;mm2 = [a3][r3][g3][b3]

		punpcklbw	mm3,mm7				;mm3 = [a4][r4][g4][b4]
		movq		mm4,mm0				;mm0 = [a1][r1][g1][b1]

		punpcklwd	mm0,mm1				;mm0 = [g2][g1][b2][b1]
		movq		mm5,mm2				;mm2 = [a3][r3][g3][b3]

		pmaddwd		mm0,[eax]
		punpcklwd	mm2,mm3				;mm2 = [g4][g3][b4][b3]

		pmaddwd		mm2,[eax+8]
		punpckhwd	mm4,mm1				;mm4 = [a2][a1][r2][r1]

		pmaddwd		mm4,[eax]
		punpckhwd	mm5,mm3				;mm5 = [a4][a3][b4][b3]

		pmaddwd		mm5,[eax+8]
		paddd		mm0,mm2				;mm0 = [ g ][ b ]

		paddd		mm0,x0000200000002000
		;idle V

		paddd		mm4,x0000200000002000
		psrad		mm0,14

		paddd		mm4,mm5				;mm4 = [ a ][ r ]

		psrad		mm4,14

		packssdw	mm0,mm4				;mm0 = [ a ][ r ][ g ][  b ]
		packuswb	mm0,mm0				;mm0 = [a][r][g][b][a][r][g][b]

		movd		eax,mm0
		ret
	}
}

static inline Pixel32 bicubic_interp_MMX(const Pixel32 *src, PixOffset pitch, unsigned long cox, unsigned long coy, const int *table) {
	Pixel32 x[4];

	cox >>= 24;
	coy >>= 24;

	src = (Pixel32 *)((char *)src - pitch - 4);

	x[0] = cc_MMX(src, table+cox*4); src = (Pixel32 *)((char *)src + pitch);
	x[1] = cc_MMX(src, table+cox*4); src = (Pixel32 *)((char *)src + pitch);
	x[2] = cc_MMX(src, table+cox*4); src = (Pixel32 *)((char *)src + pitch);
	x[3] = cc_MMX(src, table+cox*4);

	return cc_MMX(x, table + coy*4);
}

static inline Pixel32 bicubic_interp(const Pixel32 *src, PixOffset pitch, unsigned long cox, unsigned long coy, const int *table) {
	Pixel32 x[4];

	cox >>= 24;
	coy >>= 24;

	src = (Pixel32 *)((char *)src - pitch - 4);

	x[0] = cc(src, table+cox*4); src = (Pixel32 *)((char *)src + pitch);
	x[1] = cc(src, table+cox*4); src = (Pixel32 *)((char *)src + pitch);
	x[2] = cc(src, table+cox*4); src = (Pixel32 *)((char *)src + pitch);
	x[3] = cc(src, table+cox*4);

	return cc(x, table + coy*4);
}

static Pixel32 ColorRefToPixel32(COLORREF rgb) {
	return (Pixel32)(((rgb>>16)&0xff) | ((rgb<<16)&0xff0000) | (rgb&0xff00));
}

int rotate2_run(const MyFilterData2 *mfd, const VBitmap *fa_dst, const VBitmap *fa_src) {
	__int64 xaccum, yaccum;
	Pixel32 *src, *dst, pixFill;
	PixDim w, h;
	const RotateRow *rr = mfd->rows;
	const __int64 du = mfd->u_step;
	const __int64 dv = mfd->v_step;

	unsigned long Ustep, Vstep;
	unsigned long UVintstepV, UVintstepnoV;

	// initialize Abrash texmapping variables :)

	Ustep = (unsigned long)du;
	Vstep = (unsigned long)dv;

	UVintstepnoV = (long)(du>>32) + (long)(dv>>32)*(fa_src->pitch>>2);
	UVintstepV = UVintstepnoV + (fa_src->pitch>>2);

	dst = fa_dst->data;
	pixFill = ColorRefToPixel32(mfd->rgbColor);

	h = fa_dst->h;
	do {
		// texmap!

		w = rr->leftzero;
		if (w) do {
			*dst++ = pixFill;
		} while(--w);

		xaccum = rr->xaccum_left;
		yaccum = rr->yaccum_left;

		w = rr->left;
		if (w) {
			do {
				Pixel32 *src = (Pixel32*)((char *)fa_src->data + (int)(xaccum>>32)*4 + (int)(yaccum>>32)*fa_src->pitch);
				Pixel32 c1, c2, c3, c4;

				int px = (int)(xaccum >> 32);
				int py = (int)(yaccum >> 32);

				if (px<0 || py<0 || px>=fa_src->w || py>=fa_src->h)
					c1 = pixFill;
				else
					c1 = *(Pixel32 *)((char *)src + 0);

				++px;

				if (px<0 || py<0 || px>=fa_src->w || py>=fa_src->h)
					c2 = pixFill;
				else
					c2 = *(Pixel32 *)((char *)src + 4);

				++py;

				if (px<0 || py<0 || px>=fa_src->w || py>=fa_src->h)
					c4 = pixFill;
				else
					c4 = *(Pixel32 *)((char *)src + 4 + fa_src->pitch);

				--px;

				if (px<0 || py<0 || px>=fa_src->w || py>=fa_src->h)
					c3 = pixFill;
				else
					c3 = *(Pixel32 *)((char *)src + 0 + fa_src->pitch);

				*dst++ = bilinear_interp(c1, c2, c3, c4, (unsigned long)xaccum >> 24, (unsigned long)yaccum >> 24);

				xaccum += du;
				yaccum += dv;
			} while(--w);
		}

		w = fa_dst->w - (rr->leftzero + rr->left + rr->right + rr->rightzero);
		if (w) {
			__int64 xa = xaccum;
			__int64 ya = yaccum;

			src = (Pixel32*)((char *)fa_src->data + (int)(xa>>32)*4 + (int)(ya>>32)*fa_src->pitch);

			xaccum += du * w;
			yaccum += dv * w;

			if (MMX_enabled)
				do {
					*dst++ = bicubic_interp_MMX(src, fa_src->pitch, (unsigned long)xa, (unsigned long)ya, mfd->coeff_tbl);

					xa = (__int64)(unsigned long)xa + Ustep;
					ya = (__int64)(unsigned long)ya + Vstep;

					src += (xa>>32) + (ya>>32 ? UVintstepV : UVintstepnoV);

					xa = (unsigned long)xa;
					ya = (unsigned long)ya;
				} while(--w);
			else
				do {
					*dst++ = bicubic_interp(src, fa_src->pitch, (unsigned long)xa, (unsigned long)ya, mfd->coeff_tbl);

					xa = (__int64)(unsigned long)xa + Ustep;
					ya = (__int64)(unsigned long)ya + Vstep;

					src += (xa>>32) + (ya>>32 ? UVintstepV : UVintstepnoV);

					xa = (unsigned long)xa;
					ya = (unsigned long)ya;
				} while(--w);

		}

		w = rr->right;
		if (w) {
			do {
				Pixel32 *src = (Pixel32*)((char *)fa_src->data + (int)(xaccum>>32)*4 + (int)(yaccum>>32)*fa_src->pitch);
				Pixel32 c1, c2, c3, c4;

				int px = (int)(xaccum >> 32);
				int py = (int)(yaccum >> 32);

				if (px<0 || py<0 || px>=fa_src->w || py>=fa_src->h)
					c1 = pixFill;
				else
					c1 = *(Pixel32 *)((char *)src + 0);

				++px;

				if (px<0 || py<0 || px>=fa_src->w || py>=fa_src->h)
					c2 = pixFill;
				else
					c2 = *(Pixel32 *)((char *)src + 4);

				++py;

				if (px<0 || py<0 || px>=fa_src->w || py>=fa_src->h)
					c4 = pixFill;
				else
					c4 = *(Pixel32 *)((char *)src + 4 + fa_src->pitch);

				--px;

				if (px<0 || py<0 || px>=fa_src->w || py>=fa_src->h)
					c3 = pixFill;
				else
					c3 = *(Pixel32 *)((char *)src + 0 + fa_src->pitch);

				*dst++ = bilinear_interp(c1, c2, c3, c4, (unsigned long)xaccum >> 24, (unsigned long)yaccum >> 24);

				xaccum += du;
				yaccum += dv;
			} while(--w);
		}

		w = rr->rightzero;
		if (w) do {
			*dst++ = pixFill;
		} while(--w);

		dst = (Pixel32 *)((char *)dst + fa_dst->modulo);

		++rr;

	} while(--h);

	if (MMX_enabled)
		__asm emms

	return 0;
}

///////////////////////////////////////////

int rotate2_start(MyFilterData2 *mfd, int w, int h) {

	// Compute step parameters.

	double ang = mfd->angle * (3.14159265358979323846 / 2147483648.0);
	double ustep = cos(ang);
	double vstep = -sin(ang);
	__int64 du, dv;

	mfd->u_step = du = (__int64)floor(ustep*4294967296.0 + 0.5);
	mfd->v_step = dv = (__int64)floor(vstep*4294967296.0 + 0.5);

	if (!(mfd->rows = (RotateRow *)calloc(sizeof(RotateRow), h)))
		return 1;

	// It's time for Mr.Bonehead!!

	int x0, x1, x2, x3, y;
	__int64 xaccum, yaccum;
	__int64 xaccum_low, yaccum_low, xaccum_high, yaccum_high, xaccum_base, yaccum_base;
	__int64 xaccum_low2, yaccum_low2, xaccum_high2, yaccum_high2;
	RotateRow *rr = mfd->rows;

	// Compute allowable source bounds.

	xaccum_base = xaccum_low = -0x80000000i64*w;
	yaccum_base = yaccum_low = -0x80000000i64*h;
	xaccum_high = 0x80000000i64*w;
	yaccum_high = 0x80000000i64*h;

	// Compute accumulators for top-left destination position.

	xaccum = ( dv*(h-1) - du*(w-1))/2;
	yaccum = (-du*(h-1) - dv*(w-1))/2;

	// Compute 'marginal' bounds that require partial clipping.

	xaccum_low2  = xaccum_low  + 0x100000000i64;
	yaccum_low2  = yaccum_low  + 0x100000000i64;
	xaccum_high2 = xaccum_high - 0x200000000i64;
	yaccum_high2 = yaccum_high - 0x200000000i64;
	xaccum_low -= 0x100000000i64;
	yaccum_low -= 0x100000000i64;

	xaccum -= 0x80000000i64;
	yaccum -= 0x80000000i64;

	for(y=0; y<h; y++) {
		__int64 xa, ya;

		xa = xaccum;
		ya = yaccum;

		for(x0=0; x0<w; x0++) {
			if (xa >= xaccum_low && ya >= yaccum_low && xa < xaccum_high && ya < yaccum_high)
				break;

			xa += du;
			ya += dv;
		}

		rr->xaccum_left = xa - xaccum_base;
		rr->yaccum_left = ya - yaccum_base;

		for(x1=x0; x1<w; x1++) {
			if (xa >= xaccum_low2 && ya >= yaccum_low2 && xa < xaccum_high2 && ya < yaccum_high2)
				break;

			xa += du;
			ya += dv;
		}

		for(x2=x1; x2<w; x2++) {
			if (xa < xaccum_low2 || ya < yaccum_low2 || xa >= xaccum_high2 || ya >= yaccum_high2)
				break;

			xa += du;
			ya += dv;
		}

		for(x3=x2; x3<w; x3++) {
			if (xa < xaccum_low || ya < yaccum_low || xa >= xaccum_high || ya >= yaccum_high)
				break;

			xa += du;
			ya += dv;
		}

		rr->leftzero = x0;
		rr->left = x1-x0;
		rr->right = x3-x2;
		rr->rightzero = w - x3;
		++rr;

		xaccum -= dv;
		yaccum += du;
	}

	// Fill out cubic interpolation coeff. table

	if (!(mfd->coeff_tbl = (int *)malloc(sizeof(int)*256*4)))
		return 1;

	MakeCubic4Table(mfd->coeff_tbl, -0.75, !!MMX_enabled);


	return 0;
}

int rotate2_end(MyFilterData2 *mfd) {
	if ( mfd->rows )
		free(mfd->rows);
	mfd->rows = NULL;

	if ( mfd->coeff_tbl )
		free(mfd->coeff_tbl);
	mfd->coeff_tbl = NULL;
	return 0;
}