#include <windows.h>
#include <stdio.h>
#include <math.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"
#include "resample.h"
#include "cpuaccel.h"
#include <memory.h>
#include <malloc.h>
#include <fstream.h>

extern struct FilterVTbls g_vtbls;

///////////////////////////////////////////////////////////////////////////

int motionInit(FilterActivation *fa, const FilterFunctions *ff);
int tutorialRunProc(const FilterActivation *fa, const FilterFunctions *ff);
int motion_start(FilterActivation *fa, const FilterFunctions *ff);
int motion_stop(FilterActivation *fa, const FilterFunctions *ff);
long tutorialParamProc(FilterActivation *fa, const FilterFunctions *ff);
int tutorialConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void tutorialStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void tutorialScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool tutorialFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);

///////////////////////////////////////////////////////////////////////////

typedef struct RotateRow {
	int leftzero, left, right, rightzero;
	__int64 xaccum_left, yaccum_left;
} RotateRow;

typedef struct MyFilterData2 {
	int angle;
	int filtmode;

	// working variables

	IFilterPreview *ifp;

	__int64	u_step;
	__int64	v_step;

	RotateRow *rows;
	int *coeff_tbl;

	COLORREF	rgbColor;
	HBRUSH		hbrColor;

	bool	fExpandBounds;
	
} MyFilterData2;

struct MyFilterDataCopy {
	float start_zoom;
	float end_zoom;
	float start_rot;
	float end_rot;
	int x_res;
	int y_res;
	int x_start;
	int y_start;
	int x_end;
	int y_end;
	int start_frame;
	int end_frame;
	int filter_mode;
};

typedef struct MyFilterData {
	float start_zoom;
	float end_zoom;
	float start_rot;
	float end_rot;
	int x_res;
	int y_res;
	int x_start;
	int y_start;
	int x_end;
	int y_end;
	int start_frame;
	int end_frame;
	int filter_mode;
	MyFilterDataCopy undo;
	IFilterPreview *ifp;
} MyFilterData;

enum {
	FILTER_NONE				= 0,
	FILTER_BILINEAR			= 1,
	FILTER_BICUBIC			= 2,
	FILTER_TABLEBILINEAR	= 3,
	FILTER_TABLEBICUBIC075	= 4,
	FILTER_TABLEBICUBIC060	= 5,
	FILTER_TABLEBICUBIC100	= 6,
	FILTER_LANCZOS3			= 7
};

static char *filter_names[]={
	"Nearest neighbor",
	"Bilinear",
	"Bicubic",
	"Precise bilinear",
	"Precise bicubic (A=-0.75)",
	"Precise bicubic (A=-0.60)",
	"Precise bicubic (A=-1.00)",
	"Lanczos3"
};

enum {
	FILTMODE_POINT		= 0,
	FILTMODE_BILINEAR,
	FILTMODE_BICUBIC,
	FILTMODE_COUNT
};


int rotate2_run(const MyFilterData2 *mfd, const VBitmap * fa_dst, const VBitmap * fa_src);
int rotate2_start(MyFilterData2 *mfd, int w, int h);
int rotate2_end(MyFilterData2 *mfd);
void tutorialScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);

static ScriptFunctionDef motion_func_defs[]={
	{ (ScriptFunctionPtr)tutorialScriptConfig, "Config", "0iiiiiiiiiiiii" },
	{ NULL },
};

static CScriptObject motion_obj={
	NULL, motion_func_defs
};

struct FilterDefinition filterDef_tutorial = {

    NULL, NULL, NULL,       // next, prev, module
    "Motion",             // name
    "Allows pans, zooms, and rotations to be applied to the video",
                            // desc
    "Ben Greenwood",               // maker
    NULL,                   // private_data
    sizeof(MyFilterData),   // inst_data_size

    motionInit,				// initProc
    NULL,					// deinitProc
    tutorialRunProc,		// runProc
    tutorialParamProc,		// paramProc
    tutorialConfigProc,		// configProc
    tutorialStringProc,		// stringProc
    motion_start,			// startProc
    motion_stop,			// endProc

	&motion_obj,			// script_obj
    tutorialFssProc,		// fssProc

};

///////////////////////////////////////////////////////////////////////////

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_tutorial;

///////////////////////////////////////////////////////////////////////////

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd_tutorial = ff->addFilter(fm, &filterDef_tutorial, sizeof(FilterDefinition))))
        return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd_tutorial);
}

static inline void copy_sub_image_rec(unsigned char * source, unsigned char * dest, int sw, int sh, int sp, int dw, int dh, int dp, int x, int y ){
	memset(dest,0,dp*dh);
	if ( x > 0 ){
		sw-=x;
		source+=x*4;
		x=0;
	}
	if ( y > 0 ){
		sh-=y;
		source+=sp*y;
		y=0;
	}
	if ( y < 0 ){
		y=-y;
		dh-=y;
		dest+=y*dp;
		y=0;
	} 
	if ( x < 0 ){
		x=-x;
		dw-=x;
		dest+=x*4;
		x=0;
	}

	if ( dw > sw)
		dw = sw;
	if ( dh > sh)
		dh = sh;

	for ( int h =0; h < dh;h++){
		memcpy(dest+dp*h,source+sp*h,dw*4);
	}
}

void copy_sub_image_centered(unsigned char * source, unsigned char * dest, int sw, int sh, int sp, int dw, int dh, int dp, int x, int y ){
	int cx = x-dw/2;
	int cy = y-dh/2;
	copy_sub_image_rec(source,dest,sw,sh,sp,dw,dh,dp,cx,cy);

	//source rectangle is (x1,y1)-(x1+sx,y1+sy), while the destination is (x2,y2)-(x2+dx,y2+dy).
	//Init(eFilter horiz_filt, eFilter vert_filt, double dx, double dy, double sx, double sy)
	//Process(const VBitmap *dst, double _x2, double _y2, const VBitmap *src, double _x1, double _y1, bool bMapCorners) 

}

void copy_sub_image_centered2(unsigned char * source, unsigned char * dest, int sw, int sh, int sp, int dw, int dh, int dp, double x, double y, void * fa_dst, int frame){

	double cx = x-dw/2.0;
	double cy = y-dh/2.0;

	int tw = dw;
	int th = dh;
	//source rectangle is (x1,y1)-(x1+sx,y1+sy), while the destination is (x2,y2)-(x2+dx,y2+dy).
	//Init(eFilter horiz_filt, eFilter vert_filt, double dx, double dy, double sx, double sy)
	//Process(const VBitmap *dst, double _x2, double _y2, const VBitmap *src, double _x1, double _y1, bool bMapCorners) 

	float _x2=0,_y2=0;

	VFBitmap * resize_dst = (VFBitmap*)malloc(sizeof(VFBitmap));
	memcpy(resize_dst,fa_dst,sizeof(VFBitmap));
	resize_dst->data = (unsigned long*)dest;
	resize_dst->w=tw;
	resize_dst->h=th;
	resize_dst->pitch=dp;
	resize_dst->modulo=dp-(tw*4);

	VFBitmap * resize_src = (VFBitmap*)malloc(sizeof(VFBitmap));
	memcpy(resize_src,fa_dst,sizeof(VFBitmap));
	resize_src->data = (unsigned long*)source;
	resize_src->w=sw;
	resize_src->h=sh;
	resize_src->pitch=sp;
	resize_src->modulo=sp-(sw*4);

	try {
	// resize the working image
	Resampler resampler;
	//source rectangle is (x1,y1)-(x1+sx,y1+sy), while the destination is (x2,y2)-(x2+dx,y2+dy)
	//Init(eFilter horiz_filt, eFilter vert_filt, double dx, double dy, double sx, double sy)
	//Process(VBitmap *dst, double _x2, double _y2, VBitmap *src, double _x1, double _y1 )
	resampler.Init(Resampler::eFilter::kLanczos3, Resampler::eFilter::kLanczos3,dw,dh,dw,dh);
	resampler.Process(resize_dst,_x2,_y2, resize_src,cx,cy, false);
	} catch ( ... ) {}
	free(resize_dst);
	free(resize_src);
}

///////////////////////////////////////////////////////////////////////////

int tutorialRunProc(const FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData * const mfd = (MyFilterData *)fa->filter_data;
    unsigned char * const source = (unsigned char *)fa->src.data;
    unsigned char * const dest = (unsigned char *)fa->dst.data;
	const int source_height = fa->src.h;
	const int source_width = fa->src.w;
	const int source_pitch = fa->src.pitch;
	const int dest_pitch = fa->dst.pitch;
	const int dest_width = mfd->x_res;
	const int dest_height= mfd->y_res;

	// only crop video outsize the edit range
	int frame = fa->pfsi->lCurrentSourceFrame;
	if ( frame < mfd->start_frame || frame > mfd->end_frame ){
		if ( source_height == dest_height && dest_width==source_width){
			for ( int h=0;h<dest_height;h++){
				memcpy(dest+dest_pitch*h,source+source_pitch*h,source_width*4);
			}
		} else {
			copy_sub_image_centered(source,dest,source_width,source_height,source_pitch,dest_width,dest_height,dest_pitch,source_width/2,source_height/2);
		}
		return 0;
	} 
	
	frame -= mfd->start_frame;
	int frame_len = mfd->end_frame-mfd->start_frame;

	double percent = (double)frame/frame_len;

	double x = mfd->x_start;
	double y = -mfd->y_start;

	x+=((mfd->x_end-x)*percent);
	y+=(((-mfd->y_end)-y)*percent);

	// (0,0) corrisponds to the center of the source image, so convert to image
	x+= source_width/2.0;
	y+= source_height/2.0;

	double sz = mfd->start_zoom;
	double ez = mfd->end_zoom;

	// test for a pan only and apply if true
	if ( sz == ez && ez == 100 && mfd->start_rot == 0 && mfd->end_rot == 0 ){
		copy_sub_image_centered2(source,dest,source_width,source_height,source_pitch,dest_width,dest_height,dest_pitch,x,y,&(fa->dst),frame);
		return 0;
	}

	// calculate resize factors

	double curr_zoom = sz;
	curr_zoom += (ez-sz)*percent;
	curr_zoom /= 100.0;

	x*=curr_zoom;
	y*=curr_zoom;

	double zoom_width = (source_width*curr_zoom);
	double zoom_height = (source_height*curr_zoom);

	int zoom_pitch = (int)(zoom_width*4+15+4)&(~15);

	unsigned char * image = (unsigned char*)_aligned_malloc(zoom_pitch*zoom_height+32+zoom_pitch,16);

	VFBitmap * resize_dst = (VFBitmap*)malloc(sizeof(VFBitmap));
	memcpy(resize_dst,&(fa->dst),sizeof(VFBitmap));

	resize_dst->data = (unsigned long*)image;
	resize_dst->w=zoom_width+1;
	resize_dst->h=zoom_height+1;
	resize_dst->pitch=zoom_pitch;
	resize_dst->modulo=zoom_pitch-((int)zoom_width*4+4);

	// resize the working image
	Resampler resampler;
	Resampler::eFilter fmode;
	switch(mfd->filter_mode) {
		case FILTER_NONE:			fmode = Resampler::eFilter::kPoint; break;
		case FILTER_BILINEAR:		fmode = Resampler::eFilter::kLinearInterp; break;
		case FILTER_BICUBIC:		fmode = Resampler::eFilter::kCubicInterp; break;
		case FILTER_TABLEBILINEAR:	fmode = Resampler::eFilter::kLinearDecimate; break;
		case FILTER_TABLEBICUBIC060:fmode = Resampler::eFilter::kCubicDecimate060; break;
		case FILTER_TABLEBICUBIC075:fmode = Resampler::eFilter::kCubicDecimate075; break;
		case FILTER_TABLEBICUBIC100:fmode = Resampler::eFilter::kCubicDecimate100; break;
		case FILTER_LANCZOS3:		fmode = Resampler::eFilter::kLanczos3; break;
	}

	resampler.Init(fmode, fmode, zoom_width,zoom_height, source_width, source_height);
	resampler.Process(resize_dst,0,0, &(fa->src),0,0, false);

	// test for pan & zoom only, apply if true
	if ( mfd->start_rot == 0 && mfd->end_rot == 0 ){
		copy_sub_image_centered2(image,dest,zoom_width,zoom_height,zoom_pitch,dest_width,dest_height,dest_pitch,x,y,&(fa->dst),frame);
		
		_aligned_free(image);
		free(resize_dst);
		return 0;
	}
	
	double curr_rot = mfd->start_rot;
	curr_rot+=(mfd->end_rot-curr_rot)*percent;

	unsigned int d = sqrt(dest_width*dest_width+dest_height*dest_height);
	unsigned int r_pitch = (d*4+15)&(~15);

	unsigned char * image2 = (unsigned char *)_aligned_malloc(r_pitch*d,16);
	unsigned char * image3 = (unsigned char *)_aligned_malloc(r_pitch*d*4,16);

	copy_sub_image_centered2(image,image2,zoom_width,zoom_height,zoom_pitch,d,d,r_pitch,x,y,&(fa->dst),frame);

	MyFilterData2 r_data;
	r_data.angle = (int)((curr_rot*(1.0 / 360.0)-0.5) * 4294967296.0)^0x80000000;
	r_data.fExpandBounds=false;
	r_data.filtmode=FILTMODE_BICUBIC;
	r_data.rgbColor=0;

	VFBitmap * rot_dst = (VFBitmap*)malloc(sizeof(VFBitmap));
	resize_dst->pitch=r_pitch;
	resize_dst->w=d;
	resize_dst->h=d;
	resize_dst->modulo=r_pitch-(d*4);
	resize_dst->data=(unsigned long*)image2;
	memcpy(rot_dst,resize_dst,sizeof(VFBitmap));
	rot_dst->data=(unsigned long*)image3;

	rotate2_start(&r_data, d, d);
	rotate2_run(&r_data,rot_dst,resize_dst);
	rotate2_end(&r_data);

	copy_sub_image_centered(image3,dest,d,d,r_pitch,dest_width,dest_height,dest_pitch,d/2,d/2);
	
	_aligned_free(image);
	_aligned_free(image2);
	_aligned_free(image3);
	free(resize_dst);
	free(rot_dst);
	return 0;
}


long tutorialParamProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	fa->dst.pitch = (mfd->x_res*4+15)&(~15);
	fa->dst.w = mfd->x_res;
	fa->dst.h = mfd->y_res;

    return FILTERPARAM_SWAP_BUFFERS;
}

#define if_case_dlg_float_read( param, dest, low, high ) { \
	case param:\
	if (HIWORD(wParam) == EN_KILLFOCUS) { \
		float zoom; \
		int success; \
		char text[32]; \
		success = GetDlgItemText(hdlg, param, text, 32);\
		if ( success ){\
			sscanf(text,"%f",&zoom);\
			if ( zoom >= low && zoom <= high ){\
				dest=zoom;\
				mfd->ifp->RedoFrame();\
			} else {\
				success=false;\
			}\
		}\
		if (success==0 || zoom < low || zoom > high ) {\
			SetFocus((HWND)lParam);\
			MessageBeep(MB_ICONQUESTION);\
			return TRUE;\
		}\
	}\
	return TRUE;\
}

#define if_case_dlg_int_read( param, dest, low, high, is_signed, redo ) { \
	case param:\
	if (HIWORD(wParam) == EN_KILLFOCUS) { \
		int value; \
		int success; \
		value = GetDlgItemInt(hdlg, param, &success,is_signed);\
		if ( success ){\
			if ( value >= low && value <= high ){\
				dest=value;\
				if ( redo ) { \
					mfd->ifp->RedoFrame();\
				} else {\
					mfd->ifp->UndoSystem();\
					mfd->ifp->RedoSystem();\
				} \
			} else {\
				success=false;\
			}\
		}\
		if (success==0 || value < low || value > high ) {\
			SetFocus((HWND)lParam);\
			MessageBeep(MB_ICONQUESTION);\
			return TRUE;\
		}\
	}\
	return TRUE;\
}

#define move_val(x) {mfd->undo.x = mfd->x;}
#define unmove_val(x) {mfd->x = mfd->undo.x;}

#define SetDlgItemFloat( hdlg, IDC, val){\
	char text[32];\
	sprintf(text,"%.4f",val);\
	SetDlgItemText(hdlg, IDC,text);\
}

#define SetDlgItemSignedInt( hdlg, IDC, val){\
	sprintf(text,"%d",val);\
	SetDlgItemText(hdlg, IDC,text);\
}

BOOL CALLBACK tutorialConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
    MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);

    switch(msg) {
        case WM_INITDIALOG:
			{
			char text[32];
            SetWindowLong(hdlg, DWL_USER, lParam);
            mfd = (MyFilterData *)lParam;

			HWND hwndItem = GetDlgItem(hdlg, IDC_FILTER_MODE);
			
			for(int i=0; i<8; i++)
				SendMessage(hwndItem, CB_ADDSTRING, 0, (LPARAM)filter_names[i]);

			SendMessage(hwndItem, CB_SETCURSEL, mfd->undo.filter_mode, 0);
			SetDlgItemSignedInt(hdlg, IDC_WIDTH,mfd->undo.x_res);
			SetDlgItemSignedInt(hdlg, IDC_HEIGHT,mfd->undo.y_res);
			SetDlgItemSignedInt(hdlg, IDC_START_FRAME,mfd->undo.start_frame);
			SetDlgItemSignedInt(hdlg, IDC_END_FRAME,mfd->undo.end_frame);
			SetDlgItemFloat(hdlg, IDC_START_ZOOM,mfd->undo.start_zoom);
			SetDlgItemFloat(hdlg, IDC_END_ZOOM,mfd->undo.end_zoom);
			SetDlgItemFloat(hdlg, IDC_START_ROT,mfd->undo.start_rot);
			SetDlgItemFloat(hdlg, IDC_END_ROT,mfd->undo.end_rot);
			SetDlgItemSignedInt(hdlg, IDC_START_X,mfd->undo.x_start);
			SetDlgItemSignedInt(hdlg, IDC_END_X,mfd->undo.x_end,);
			SetDlgItemSignedInt(hdlg, IDC_START_Y,mfd->undo.y_start);
			SetDlgItemSignedInt(hdlg, IDC_END_Y,mfd->undo.y_end);

			mfd->ifp->InitButton(GetDlgItem(hdlg, IDC_PREVIEW));
			
			}

            return TRUE;

        case WM_COMMAND:
            switch(LOWORD(wParam)) {
				if_case_dlg_float_read( IDC_START_ZOOM, mfd->start_zoom, 10, 512 );
				if_case_dlg_float_read( IDC_END_ZOOM, mfd->end_zoom, 10, 512 );
				if_case_dlg_float_read( IDC_START_ROT, mfd->start_rot, -10*1024, 10*1024 );
				if_case_dlg_float_read( IDC_END_ROT, mfd->end_rot, -10*1024, 10*1024 );
				if_case_dlg_int_read( IDC_WIDTH, mfd->x_res, 16, 1024*4, false, true );
				if_case_dlg_int_read( IDC_HEIGHT, mfd->y_res, 16, 1024*4, false, true );
				if_case_dlg_int_read( IDC_START_X, mfd->x_start, -2048, 2048, true, false );
				if_case_dlg_int_read( IDC_START_Y, mfd->y_start, -2048, 2048, true, false );
				if_case_dlg_int_read( IDC_END_X, mfd->x_end, -2048, 2048, true, false );
				if_case_dlg_int_read( IDC_END_Y, mfd->y_end, -2048, 2048, true, false );
				if_case_dlg_int_read( IDC_START_FRAME, mfd->start_frame, 0, 0x7FFFFFFF, false, false );
				if_case_dlg_int_read( IDC_END_FRAME, mfd->end_frame, 0, 0x7FFFFFFF, false, false );
			case IDC_FILTER_MODE:
				mfd->filter_mode = SendDlgItemMessage(hdlg, IDC_FILTER_MODE, CB_GETCURSEL, 0, 0);
				mfd->ifp->RedoFrame();
				return TRUE;

			case IDC_PREVIEW:
				mfd->ifp->Toggle(hdlg);
				return TRUE;

            case IDOK:

				move_val(start_zoom);
				move_val(end_zoom);
				move_val(start_rot);
				move_val(end_rot);
				move_val(x_res);
				move_val(y_res);
				move_val(x_start);
				move_val(y_start);
				move_val(x_end);
				move_val(y_end);
				move_val(start_frame);
				move_val(end_frame);
				move_val(filter_mode);

				mfd->ifp->Close();

                EndDialog(hdlg, 0);
                return TRUE;
            case IDCANCEL:

				unmove_val(start_zoom);
				unmove_val(end_zoom);
				unmove_val(start_rot);
				unmove_val(end_rot);
				unmove_val(x_res);
				unmove_val(y_res);
				unmove_val(x_start);
				unmove_val(y_start);
				unmove_val(x_end);
				unmove_val(y_end);
				unmove_val(start_frame);
				unmove_val(end_frame);
				unmove_val(filter_mode);
				mfd->ifp->Close();
                EndDialog(hdlg, 1);
                return FALSE;
            }
            break;
    }

    return FALSE;
}

int tutorialConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
    mfd->ifp = fa->ifp;
	return DialogBoxParam(fa->filter->module->hInstModule,
            MAKEINTRESOURCE(IDD_FILTER_TUTORIAL), hwnd,
            tutorialConfigDlgProc, (LPARAM)fa->filter_data);
}

void tutorialStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	_snprintf(str,80," - Frames %d-%d",mfd->start_frame,mfd->end_frame);
}
void tutorialScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
    FilterActivation *fa = (FilterActivation *)lpVoid;
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->x_res		=argv[0].asInt();
	mfd->y_res		=argv[1].asInt();
	mfd->x_start	=argv[2].asInt();
	mfd->x_end		=argv[3].asInt();
	mfd->y_start	=argv[4].asInt();
	mfd->y_end		=argv[5].asInt();
	mfd->start_frame=argv[6].asInt();
	mfd->end_frame	=argv[7].asInt();
	mfd->start_zoom =argv[8].asInt()/1000.0;
	mfd->end_zoom	=argv[9].asInt()/1000.0;
	mfd->start_rot	=argv[10].asInt()/1000.0;
	mfd->end_rot	=argv[11].asInt()/1000.0;
	mfd->filter_mode=argv[12].asInt();

	memcpy(&(mfd->undo),mfd,sizeof(mfd->undo));
}

bool tutorialFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	_snprintf(buf, buflen, "Config(%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d)",
		mfd->x_res, mfd->y_res, mfd->x_start,mfd->x_end, mfd->y_start, mfd->y_end,
		mfd->start_frame, mfd->end_frame,
		(int)(mfd->start_zoom*1000),(int)(mfd->end_zoom*1000),
		(int)(mfd->start_rot*1000),(int)(mfd->end_rot*1000),
		mfd->filter_mode );
		
    return true;
}

static int motion_start(FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	return 0;
}

static int motion_stop(FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	return 0;
}

static int motionInit(FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;

	mfd->undo.start_zoom=100;
	mfd->undo.end_zoom=100;
	mfd->undo.start_rot=0;
	mfd->undo.end_rot=0;
	mfd->undo.x_res=640;
	mfd->undo.y_res=480;
	mfd->undo.x_start=0;
	mfd->undo.y_start=0;
	mfd->undo.x_end=0;
	mfd->undo.y_end=0;
	mfd->undo.start_frame=0;
	mfd->undo.end_frame=100;
	mfd->undo.filter_mode=FILTER_BICUBIC;

	mfd->start_zoom=100;
	mfd->end_zoom=100;
	mfd->start_rot=0;
	mfd->end_rot=0;
	mfd->x_res=640;
	mfd->y_res=480;
	mfd->x_start=0;
	mfd->y_start=0;
	mfd->x_end=0;
	mfd->y_end=0;
	mfd->start_frame=0;
	mfd->end_frame=100;
	mfd->filter_mode=FILTER_BICUBIC;

	return 0;
}

//MessageBox (HWND_DESKTOP, msg, "Error", MB_OK | MB_ICONEXCLAMATION);